#  tests for pyfftw-0.11.1-py36h452e1ab_1000 (this is a generated file);
print('===== testing package: pyfftw-0.11.1-py36h452e1ab_1000 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import numpy as np

import pyfftw.builders
from pyfftw.interfaces.numpy_fft import fftn
r = np.random.randn(32, 32, 32)

# the following transform will fail if MKL FFT routines are being linked to
# instead of the FFTW ones.  (MKL doesn't support FFT of only 1
# dimension of a 3D array).
# see:  https://github.com/pyFFTW/pyFFTW/issues/40
fftn(r, axes=(0, ))
#  --- run_test.py (end) ---

print('===== pyfftw-0.11.1-py36h452e1ab_1000 OK =====');
